#ifndef CURL_EASY_H
#define CURL_EASY_H

#include <string>
#include <vector>
#include <curl.h>
#include "sharedPtr.h"


/**
 * Encapsulates the cURL library easy interface in an exception safe manner.
 */
class CURLEasy {

public:

	/** Default exception. */
	class Exception {};


	/** Adaption of "curl_easy_escape". */
	class Escaped {

	public:

		Escaped(const std::string& str);
		~Escaped();

		std::string get() const;

	private:

		char* buffer;

	};


	/**
	 * Adaption of a linked list of "curl_httppost" for http posts.
	 */
	class Form {
		
	public:

		Form();
		virtual ~Form();

		/** Returns a pointer to the form which may be used in "setOpt". */
		struct curl_httppost* get() const;

		/**
		 * Macro expands to a function call that adds a form to the list. Its
		 * first parameter must be an object of this class. Further parameters
		 * are passed as the variadic arguments to "curl_formadd".
		 */
		#define CURL_EASY_FORM_ADD(form,...) \
			if (curl_formadd(&(form).first,&(form).last,__VA_ARGS__) != CURLE_OK) { \
				throw ::CURLEasy::Exception(); \
			}


		struct curl_httppost* first;
		struct curl_httppost* last;

	private:

		/** Copying and assigning prohibited. */
		Form(const Form& obj);
		Form& operator=(const Form& rhs);

	};


	CURLEasy();
	CURLEasy(const CURLEasy& obj);
	~CURLEasy();

	CURLEasy& operator=(CURLEasy obj);
	void swap(CURLEasy& obj);

	/** Adaption of "curl_easy_setopt". */
	template <typename T>
	void setOpt(CURLoption option, T param);

	/** Adaption of "curl_easy_getinfo". */
	template <typename T>
	void getInfo(CURLINFO info, T param) const;

	/** Adaption of "curl_easy_perform". */
	void perform();

private:

	/**
	 * Calls the global initialization and cleanup functions of the cURL
	 * library.
	 */
	class Global {

	public:

		Global();
		~Global();

	};


	CURL* handle;


	static const Global init;

};



template <typename T>
void CURLEasy::setOpt(CURLoption option, T param) {
	if (curl_easy_setopt(handle,option,param) != CURLE_OK) {
		throw Exception();
	}
}

template <typename T>
void CURLEasy::getInfo(CURLINFO info, T param) const {
	if (curl_easy_getinfo(handle,info,param) != CURLE_OK) {
		throw Exception();
	}
}

#endif