#ifndef DERIVED_H
#define DERIVED_H


/**
 * Provides by-value semantics for polymorphism. An object of this class with
 * template parameter T represents an object of any derived class of T
 * (including T itself). In fact, just the pointer conversion is required.
 */
template <typename T>
class Derived {

public:

	/**
	 * Template parameter U must be a derived class of T, or at least U* must
	 * be convertible to T*.
	 */
	template <typename U>
	Derived(const U& object);

	Derived(const Derived& src);
	~Derived();

	Derived<T>& operator=(Derived rhs);
	void swap(Derived& rhs);

	operator T*() const;
	T* operator->() const;

private:

	class ObjectBase {
		
	public:

		virtual ~ObjectBase();

		virtual T* get() = 0;
		virtual ObjectBase* copy() const = 0;

	};


	template <typename U>
	class Object : public ObjectBase {

	public:

		Object(const U& object);

		T* get();
		ObjectBase* copy() const;

	private:

		U object;

	};


	ObjectBase* object;

};



template <typename T>
template <typename U>
Derived<T>::Derived(const U& object) : object(new Object<U>(object)) {}

template <typename T>
Derived<T>::Derived(const Derived& src) : object(src.object->copy()) {}

template <typename T>
Derived<T>::~Derived() {
	delete object;
}

template <typename T>
Derived<T>& Derived<T>::operator=(Derived rhs) {
	swap(rhs);
	return *this;
}

template <typename T>
void Derived<T>::swap(Derived& rhs) {
	std::swap(object,rhs.object);
}

template <typename T>
Derived<T>::operator T*() const {
	return object->get();
}

template <typename T>
T* Derived<T>::operator->() const {
	return object->get();
}

template <typename T>
Derived<T>::ObjectBase::~ObjectBase() {}

template <typename T>
template <typename U>
Derived<T>::Object<U>::Object(const U& object) : object(object) {}

template <typename T>
template <typename U>
T* Derived<T>::Object<U>::get() {
	return &object;
}

template <typename T>
template <typename U>
typename Derived<T>::ObjectBase* Derived<T>::Object<U>::copy() const {
	return new Object<U>(object);
}

#endif