#ifndef DESTINATION_H
#define DESTINATION_H

#include <string>
#include "derived.h"
#include "thread.h"
#include "sharedObject.h"
#include "file.h"


/**
 * Derived classes represent output targets byte sequence data. An object of
 * such a class actually just decribes the location. To output data, an object
 * of the nested class "Writer" is needed. This is returned by "getWriter"
 * which has a member to write to the destination.
 */
class Destination {

public:

	/**
	 * Default exception for this class.
	 */
	class Exception {};


	/**
	 * Its constructor and destructor are supposed to do open/connect or
	 * close/disconnect operations, respectively.
	 */
	class Writer {

	public:

		virtual ~Writer();

		/**
		 * Writes "data" to the destination. It must contain "size" bytes.
		 */
		virtual void write(const char* data, size_t size) = 0;

	};


	virtual ~Destination();

	/**
	 * Returns a writer object for this destination.
	 */
	virtual Derived<Writer> getWriter() = 0;

};


/**
 * A buffer is used as output target.
 */
class DestinationBuffer : public Destination {

public:

	/**
	 * Corresponding writer.
	 */
	class Writer : public Destination::Writer {

	public:

		/**
		 * Construction of a writer will clear the buffer. Copying and
		 * assigning will not.
		 */
		Writer(const Thread::SharedPtr<SharedObject<std::string> >& buffer);
		~Writer();

		void write(const char* data, size_t size);

	private:

		Thread::SharedPtr<SharedObject<std::string> > buffer;
		std::string bufferCopy;

	};


	DestinationBuffer(
		const Thread::SharedPtr<SharedObject<std::string> >& buffer);

	Derived<Destination::Writer> getWriter();

private:

	Thread::SharedPtr<SharedObject<std::string> > buffer;

};


/**
 * A file is used as output target.
 */
class DestinationFile : public Destination {

public:

	/**
	 * Corresponding writer. Note that copied writers refer to the same file
	 * handle and therefore will share the same position indicator.
	 */
	class Writer : public Destination::Writer {

	public:

		Writer(const std::string& file);

		void write(const char* data, size_t size);

	private:

		SharedPtr<File> handle;

	};


	DestinationFile(const std::string& file);

	Derived<Destination::Writer> getWriter();

private:

	std::string file;

};

#endif