#include <string>
#include "httpRequest.h"
#include "httpTransfer.h"
#include "destination.h"
#include "source.h"
#include "derived.h"
#include "sharedPtr.h"
#include "gmid.h"

#define _DLL_H_
#define export extern "C" __declspec (dllexport)


/** HTTP request wrapper struct for use in Game Maker. */
struct HTTPRequestGM {

	/** Using the buffer as default destination. */
	HTTPRequestGM() : buffer(new SharedObject<std::string>),
		request(DestinationBuffer(buffer)), transfer(0) {}

	~HTTPRequestGM() {
		delete transfer;
	}

	Thread::SharedPtr<SharedObject<std::string> > buffer;
	std::string bufferCopy;

	HTTPRequest request;
	HTTPTransfer* transfer;

};


/**
 * Mapping between request objects and integer identifiers.
 */
GMID<SharedPtr<struct HTTPRequestGM> > httpRequestSet;


/**
 * Creates a new http request and returns its id or -1 in case of failure.
 */
export double http_request_create() {
	try {
		return httpRequestSet.add(
			SharedPtr<struct HTTPRequestGM>(new HTTPRequestGM));
	} catch (...) {
		return -1;
	}
}

/**
 * Destroys a given request. It returns whether the given id was valid.
 */
export double http_request_destroy(double id) {
	try {
		httpRequestSet.remove(id);
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Sets the URL for a given request.
 */
export double http_request_set_url(double id, const char* url) {
	try {
		httpRequestSet.get(id)->request.setURL(url);
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Sets the given file as output target.
 */
export double http_request_set_file(double id, const char* file) {
	try {
		httpRequestSet.get(id)->request.setDestination(DestinationFile(file));
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Sets the internal buffer of a request a output target. This is the default.
 */
export double http_request_set_buffer(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		ptr->request.setDestination(DestinationBuffer(ptr->buffer));
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * The following four functions add a name-value-pair for submission via GET or
 * POST with its value taken from a string or file.
 */
export double http_request_add_get_string(double id, const char* name, const char* value) {
	try {
		httpRequestSet.get(id)->request.getListGET().push_back(
			HTTPRequest::Pair(SourceBuffer(name),SourceBuffer(value)));
	} catch (...) {
		return 0;
	}
	return 1;
}

export double http_request_add_get_file(double id, const char* name, const char* file) {
	try {
		httpRequestSet.get(id)->request.getListGET().push_back(
			HTTPRequest::Pair(SourceBuffer(name),SourceFile(file)));
	} catch (...) {
		return 0;
	}
	return 1;
}

export double http_request_add_post_string(double id, const char* name, const char* value) {
	try {
		httpRequestSet.get(id)->request.getListPOST().push_back(
			HTTPRequest::Pair(SourceBuffer(name),SourceBuffer(value)));
	} catch (...) {
		return 0;
	}
	return 1;
}

export double http_request_add_post_file(double id, const char* name, const char* file) {
	try {
		httpRequestSet.get(id)->request.getListPOST().push_back(
			HTTPRequest::Pair(SourceBuffer(name),SourceFile(file)));
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * The following two functions add file uploads to the request. The data is
 * either taken from a string or read from a file. The second parameter is the
 * filename that is actually sent and may differ from the filename of the given
 * file in the second function.
 */
export double http_request_add_upload_string(double id, const char* name, const char* filename, const char* data) {
	try {
		httpRequestSet.get(id)->request.getListUPLOAD().push_back(
			HTTPRequest::Upload(SourceBuffer(name),SourceBuffer(filename),SourceBuffer(data)));
	} catch (...) {
		return 0;
	}
	return 1;
}

export double http_request_add_upload_file(double id, const char* name, const char* filename, const char* file) {
	try {
		httpRequestSet.get(id)->request.getListUPLOAD().push_back(
			HTTPRequest::Upload(SourceBuffer(name),SourceBuffer(filename),SourceFile(file)));
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Clearing all the data that was added for submission via GET.
 */
export double http_request_clear_get(double id) {
	try {
		httpRequestSet.get(id)->request.getListGET().clear();
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Clearing all the data that was added for submission via POST.
 */
export double http_request_clear_post(double id) {
	try {
		httpRequestSet.get(id)->request.getListPOST().clear();
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Clearing all the data that was added for file uploads.
 */
export double http_request_clear_upload(double id) {
	try {
		httpRequestSet.get(id)->request.getListUPLOAD().clear();
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Performs a given request in a new thread, i.e. this function does not wait
 * to return until the transfer is finished.
 */
export double http_request_perform(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);

		if (ptr->transfer == 0 ||
			!ptr->transfer->getPerforming()) {

			HTTPTransfer* newTransfer = new HTTPTransfer(ptr->request);
			delete ptr->transfer;
			ptr->transfer = newTransfer;
		}
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Waits for the currently running transfer to finish.
 */
export double http_request_wait(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			ptr->transfer->wait();
		}
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Aborts a currently performing request.
 */
export double http_request_abort(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			ptr->transfer->abort();
		}
	} catch (...) {
		return 0;
	}
	return 1;
}

/**
 * Progression functions.
 */
export double http_request_get_down_total(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			return ptr->transfer->getDownTotal();
		}
	} catch (...) {}
	return -1;
}

export double http_request_get_down_current(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			return ptr->transfer->getDownCurrent();
		}
	} catch (...) {}
	return -1;
}

export double http_request_get_up_total(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			return ptr->transfer->getUpTotal();
		}
	} catch (...) {}
	return -1;
}

export double http_request_get_up_current(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			return ptr->transfer->getUpCurrent();
		}
	} catch (...) {}
	return -1;
}

/**
 * Returns whether the given request is still/already in progress.
 */
export double http_request_get_performing(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			return ptr->transfer->getPerforming();
		}
	} catch (...) {}
	return 0;
}

/**
 * Returns whether the given request is still/already performing.
 */
export double http_request_get_successful(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			return ptr->transfer->getSuccessful();
		}
	} catch (...) {}
	return 0;
}

/**
 * Returns the http status code of the last successful request.
 */
export double http_request_get_status_code(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		if (ptr->transfer != 0) {
			return ptr->transfer->getHTTPCode();
		}
	} catch (...) {}
	return 0;
}

/**
 * Retrieves the contents of the target buffer of a given request as a string.
 */
export const char* http_request_get_buffer(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		{
			SharedObject<std::string>::Lock lock(*ptr->buffer);
			ptr->bufferCopy = lock.get();
		}
		return ptr->bufferCopy.c_str();
	} catch (...) {}
	return "";
}

/**
 * Retrieves the size of the target buffer of a given request.
 */
export double http_request_get_buffer_size(double id) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		SharedObject<std::string>::Lock lock(*ptr->buffer);
		return lock.get().size();
	} catch (...) {}
	return -1;
}

/**
 * Retrieves a particular character of the target buffer of a given request.
 */
export double http_request_get_buffer_char(double id, double index) {
	try {
		SharedPtr<struct HTTPRequestGM> ptr = httpRequestSet.get(id);
		unsigned int i = static_cast<unsigned int>(index);
		SharedObject<std::string>::Lock lock(*ptr->buffer);
		if (i < lock.get().size()) {
			return static_cast<unsigned char>(lock.get()[i]);
		}
	} catch (...) {}
	return -1;
}

/*struct TestBuffer {
	char* data;
	size_t size;
};

size_t testRead(void *ptr, size_t size, size_t nmemb, void *userdata) {
	TestBuffer* buffer = (TestBuffer*)userdata;
	
	size_t total = size*nmemb;
	size_t read_size = total > buffer->size ? buffer->size : total;
	
	memcpy(ptr,buffer->data,read_size);

	buffer->data += read_size;
	buffer->size -= read_size;

	return read_size;
}*/

//export const char* test() {
	/*std::string* buffer = new std::string;
	HTTPRequest request = HTTPRequest(DestinationBuffer(*buffer));
	request.setURL("http://blackspark.bplaced.net/test.php");

	request.getListGET().push_back(
		HTTPRequest::Pair(SourceBuffer("name1*()"),SourceBuffer("value1")));
	request.getListGET().push_back(
		HTTPRequest::Pair(SourceBuffer("name2"),SourceBuffer("value2")));

	request.getListPOST().push_back(
		HTTPRequest::Pair(SourceBuffer("name3*()"),SourceBuffer("value3")));
	request.getListPOST().push_back(
		HTTPRequest::Pair(SourceBuffer("name4"),SourceBuffer("value4")));

	request.getListUPLOAD().push_back(
		HTTPRequest::Upload(SourceBuffer("upload1"),SourceBuffer("file1"),SourceFile("test.txt")));

	HTTPTransfer transfer(request);
	transfer.wait();

	return buffer->c_str();*/


	/*SourceBuffer src1("hallo");
	SourceFile src2("test.txt");
	SourceURLEncoded src3(SourceBuffer("123=123:__"));

	std::vector<Derived<Source> > concat;
	concat.push_back(SourceFile("test.txt"));
	concat.push_back(SourceBuffer("wie gehts?"));
	SourceConcat src4(concat);

	std::vector<std::pair<Derived<Source>,Derived<Source> > > query;
	query.push_back(std::pair<Derived<Source>,Derived<Source> >(SourceBuffer("name1"),SourceBuffer("*!)value1")));
	query.push_back(std::pair<Derived<Source>,Derived<Source> >(SourceBuffer("name2"),SourceBuffer("value2")));
	SourceURLQuery src5(query);
	
	std::string* buffer = new std::string(src5.getReader()->toString());
	return buffer->c_str();*/


	/*char buffer[1000];
	File handle("test.txt","rb");
	handle.read(buffer,handle.size());

	std::string* bufferRet = new std::string(buffer,handle.size());
	return bufferRet->c_str();*/


	/*SharedPtr<int> testVar(new int(100));
	SharedPtr<int> testVar2(new int(200));
	testVar = testVar2;

	*testVar = 0;
	*testVar2 = 0;

	return "";*/


	/*Derived<int> testVar(int(100));
	Derived<int> testVar2(testVar);
	testVar2 = testVar;

	*testVar2 = 0;

	return "";*/


	/*struct TestBuffer input = {"data",4};

	struct curl_httppost* first = 0;
	struct curl_httppost* last = 0;

	curl_formadd(&first,&last,
		CURLFORM_COPYNAME,"name",
		CURLFORM_STREAM,&input,
		CURLFORM_CONTENTSLENGTH,input.size,
		CURLFORM_END);

	CURL* curl = curl_easy_init();
	FILE* output = fopen("output.txt","wb");

	curl_easy_setopt(curl,CURLOPT_URL,"http://blackspark.bplaced.net/test.php");
	curl_easy_setopt(curl,CURLOPT_WRITEDATA,output);
	curl_easy_setopt(curl,CURLOPT_READFUNCTION,testRead);
	curl_easy_setopt(curl,CURLOPT_HTTPPOST,first);

	curl_easy_perform(curl);

	fclose(output);
	curl_formfree(first);

	return "";*/

	/*SourceFile src("test.txt");
	Derived<Source::Reader> reader(src.getReader());*/

//	return "";
//}