#ifndef GMID_H
#define GMID_H

#include <map>


/**
 * Facilitates the use of integer identifiers for objects when developing for
 * Game Maker.
 */
template <typename T>
class GMID {
	
public:

	/**
	 * Exception type thrown by members below.
	 */
	class Exception {};


	/**
	 * Adds the given object and returns the assigned id. It may throw an
	 * exception in case of failure.
	 */
	double add(const T& init = T());

	/**
	 * Removes the object with the given id. An exception is thrown if an
	 * invalid or non-existent id is given.
	 */
	void remove(double id);

	/**
	 * Removes all objects.
	 */
	void clear();

	/**
	 * Retrieves the object with the given id. If the id is invalid or
	 * non-existent an exception is thrown.
	 */
	T& get(double id);
	
private:

	std::map<unsigned int,T> mapID;

};



template <typename T>
double GMID<T>::add(const T& init) {
	unsigned int newID = 0;

	if (!mapID.empty()) {
		newID = mapID.rbegin()->first+1;
	}

	mapID.insert(std::pair<unsigned int,T>(newID,init));
	return newID;
}

template <typename T>
void GMID<T>::remove(double id) {
	if (id < 0) {
		throw Exception();
	}
	
	unsigned int int_id = static_cast<unsigned int>(id);

	if (mapID.erase(int_id) == 0) {
		throw Exception();
	}
}

template <typename T>
void GMID<T>::clear() {
	mapID.clear();
}

template <typename T>
T& GMID<T>::get(double id) {
	if (id < 0) {
		throw Exception();
	}
	
	unsigned int int_id = static_cast<unsigned int>(id);

	std::map<unsigned int,T>::iterator it = mapID.find(int_id);

	if (it == mapID.end()) {
		throw Exception();
	}

	return it->second;
}

#endif