#ifndef HTTP_REQUEST_H
#define HTTP_REQUEST_H

#include <string>
#include <vector>
#include "derived.h"
#include "destination.h"
#include "curlEasy.h"
#include "source.h"


/**
 * Represents a single http request. It uses the cURL library so make sure
 * "curl_global_init" has been called before using this class.
 */
class HTTPRequest {

public:

	/** Exception type for this class. */
	class Exception {};


	/**
	 * Represents a single name-value-pair for submission via GET or POST.
	 */
	class Pair {

	public:

		Pair(const Derived<Source>& name, const Derived<Source>& value);

		const Derived<Source>& getName() const;
		const Derived<Source>& getValue() const;

	private:

		Derived<Source> name;
		Derived<Source> value;

	};


	/**
	 * Represents a single file for file upload.
	 */
	class Upload {

	public:

		Upload(const Derived<Source>& name,
			const Derived<Source>& filename,
			const Derived<Source>& data);

		const Derived<Source>& getName() const;
		const Derived<Source>& getFilename() const;
		const Derived<Source>& getData() const;

	private:

		Derived<Source> name;
		Derived<Source> filename;
		Derived<Source> data;

	};


	/**
	 * Provides the url encoding of another source.
	 */
	class SourceURLEncoded : public Source {
		
	public:

		/**
		 * Corresponding reader.
		 */
		class Reader : public Source::Reader {
			
		public:

			Reader(const Derived<Source>& src);

			size_t read(char* data, size_t numb);
			void rewind();

		private:

			Derived<Source::Reader> reader;
			std::string buffer;
			size_t bufferPos;

		};


		SourceURLEncoded(const Derived<Source>& src);

		Derived<Source::Reader> getReader();

	private:

		Derived<Source> src;

	};


	/**
	 * Provides an url query out of a list of name-value-pairs. i.e. each name is
	 * separated by "=" from its value, the pairs again are separated by "&" and
	 * each name and value is url encoded.
	 */
	class SourceURLQuery : public SourceConcat {

	public:

		/**
		 * Each element in the given vector represents a name-value-pair. Within
		 * each pair, the first element holds the name and the second element the
		 * corresponding value.
		 */
		SourceURLQuery(const std::vector<HTTPRequest::Pair>& src);

	private:
		
		std::vector<Derived<Source> > concat(
			const std::vector<HTTPRequest::Pair>& src);

	};


	/** Represents the URL of a given request. */
	class SourceURL : public SourceConcat {

	public:

		SourceURL(const HTTPRequest& request);

	private:

		std::vector<Derived<Source> > concat(const HTTPRequest& request);

	};


	/** Represents the POST query of a given request. */
	class SourcePOST : public SourceURLQuery {

	public:

		SourcePOST(const HTTPRequest& request);

	};


	/**
	 * Represents a form including all the POST data and file uploads of the
	 * given request.
	 */
	class Form : public CURLEasy::Form {

	public:

		Form(const HTTPRequest& request);

	private:

		std::vector<Derived<Source::Reader> > readerPOST;
		std::vector<Derived<Source::Reader> > readerUPLOAD;


		void addPOST(const HTTPRequest& request);
		void addUPLOAD(const HTTPRequest& request);

		/** Copying and assigning prohibited. */
		Form(const Form& obj);
		Form& operator=(const Form& rhs);

	};


	HTTPRequest(const Derived<Destination>& dest);

	/** Sets the URL used for the upcoming request. */
	void setURL(const std::string& url);

	/** Retrieves the current URL. */
	const std::string& getURL() const;

	/** Sets the output target as an object of "Destination". */
	void setDestination(const Derived<Destination>& dest);

	/** Retrieves the current destination. */
	const Derived<Destination>& getDestination() const;

	/** List of name-value-pairs that are appended to the url. */
	std::vector<Pair>& getListGET();
	const std::vector<Pair>& getListGET() const;

	/** List of name-value-pairs that are sent via POST. */
	std::vector<Pair>& getListPOST();
	const std::vector<Pair>& getListPOST() const;

	/** List of file uploads. */
	std::vector<Upload>& getListUPLOAD();
	const std::vector<Upload>& getListUPLOAD() const;

private:

	std::string url;

	/** Output target for received data. */
	Derived<Destination> dest;

	/** Lists holding data for GET and POST submission. */
	std::vector<Pair> listGET;
	std::vector<Pair> listPOST;

	/** List holding file uploads. */
	std::vector<Upload> listUPLOAD;

};

#endif