#include "httpTransfer.h"
#include "source.h"


HTTPTransfer::HTTPTransfer(const HTTPRequest& request) :
	shared(new SharedData), thread(Runner(request,shared)) {}

HTTPTransfer::~HTTPTransfer() {
	abort();
}

double HTTPTransfer::getDownTotal() {
	SharedObject<SharedData::Progress>::Lock lock(shared->progress);
	return lock.get().dltotal;
}

double HTTPTransfer::getDownCurrent() {
	SharedObject<SharedData::Progress>::Lock lock(shared->progress);
	return lock.get().dlnow;
}

double HTTPTransfer::getUpTotal() {
	SharedObject<SharedData::Progress>::Lock lock(shared->progress);
	return lock.get().ultotal;
}

double HTTPTransfer::getUpCurrent() {
	SharedObject<SharedData::Progress>::Lock lock(shared->progress);
	return lock.get().ulnow;
}

void HTTPTransfer::abort() const {
	SharedObject<bool>::Lock lock(shared->aborting);
	lock.get() = true;
}

void HTTPTransfer::wait() const {
	thread.wait();
}

bool HTTPTransfer::getPerforming() const {
	return thread.wait(0);
}

bool HTTPTransfer::getSuccessful() const {
	return thread.successful();
}

long HTTPTransfer::getHTTPCode() const {
	SharedObject<long>::Lock lock(shared->httpCode);
	return lock.get();
}

int HTTPTransfer::callbackProgress(
	void *clientp, double dltotal, double dlnow, double ultotal, double ulnow) {

	Runner* runner = reinterpret_cast<Runner*>(clientp);

	{
		SharedObject<SharedData::Progress>::Lock
			lock(runner->shared->progress);

		lock.get().dltotal = dltotal;
		lock.get().dlnow = dlnow;
		lock.get().ultotal = ultotal;
		lock.get().ulnow = ulnow;
	}

	SharedObject<bool>::Lock lock(runner->shared->aborting);
	return lock.get();
}

size_t HTTPTransfer::callbackRead(
	void *ptr, size_t size, size_t nmemb, void *userdata) {

	Source::Reader* reader = reinterpret_cast<Source::Reader*>(userdata);
	char* data = reinterpret_cast<char*>(ptr);

	try {
		return reader->read(data,size*nmemb);
	} catch (...) {
		return CURL_READFUNC_ABORT;
	}
}

size_t HTTPTransfer::callbackWrite(
	void *ptr, size_t size, size_t nmemb, void *userdata) {

	Destination::Writer* writer =
		reinterpret_cast<Destination::Writer*>(userdata);
	char* data = reinterpret_cast<char*>(ptr);
	size_t total = size*nmemb;

	try {
		writer->write(data,total);
	} catch (...) {
		return 0;
	}
	return total;
}

HTTPTransfer::SharedData::SharedData() :
	aborting(false), httpCode(0) {}

HTTPTransfer::Runner::Runner(const HTTPRequest& request,
	const Thread::SharedPtr<SharedData>& shared) :
		request(request), shared(shared) {}

void HTTPTransfer::Runner::run() {
	CURLEasy curl;

	// Basic setup for cURL.
	curl.setOpt(CURLOPT_PROTOCOLS,CURLPROTO_HTTP);
	curl.setOpt(CURLOPT_WRITEFUNCTION,callbackWrite);
	curl.setOpt(CURLOPT_READFUNCTION,callbackRead);
	curl.setOpt(CURLOPT_PROGRESSFUNCTION,callbackProgress);
	curl.setOpt(CURLOPT_PROGRESSDATA,this);
	curl.setOpt(CURLOPT_NOPROGRESS,0);

	// Create URL.
	HTTPRequest::SourceURL sourceURL(request);
	std::string url(sourceURL.getReader()->toString());

	// Create and open source for POST query.
	HTTPRequest::SourcePOST sourcePOST(request);
	Derived<Source::Reader> readerPOST(sourcePOST.getReader());

	// Create form data for possible file uploads.
	HTTPRequest::Form form(request);

	// Open destination.
	Derived<Destination::Writer> writer(
		request.getDestination()->getWriter());

	// Detailed setup for cURL.
	curl.setOpt(CURLOPT_URL,url.c_str());
	curl.setOpt(CURLOPT_WRITEDATA,(Destination::Writer*)(writer));
	curl.setOpt(CURLOPT_READDATA,(Source::Reader*)(readerPOST));

	if (!request.getListUPLOAD().empty()) {
		// Includes file uploads.
		curl.setOpt(CURLOPT_HTTPPOST,form.get());
	} else if (!request.getListPOST().empty()) {
		// No file uploads but data to send via POST.
		curl.setOpt(CURLOPT_POST,1);
		curl.setOpt(CURLOPT_POSTFIELDS,0);
		curl.setOpt(CURLOPT_POSTFIELDSIZE,readerPOST->size());
	} // else: Neither file uploads nor POST data. Perform normal GET.

	// Perform request
	curl.perform();

	// Update http code.
	{
		SharedObject<long>::Lock lock(shared->httpCode);
		curl.getInfo(CURLINFO_RESPONSE_CODE,&lock.get());
	}
}