#ifndef HTTP_TRANSFER_H
#define HTTP_TRANSFER_H

#include "httpRequest.h"
#include "sharedObject.h"
#include "thread.h"


/**
 * Performs a given request. The passed request object is copied, so that
 * the original one may be modified while the transfer is still in
 * progress. The transfer starts immediately after construction.
 */
class HTTPTransfer {

public:

	HTTPTransfer(const HTTPRequest& request);

	/** Automatically aborts the transfer but waits for it to finish. */
	~HTTPTransfer();

	/** Returns progression information. */
	double getDownTotal();
	double getDownCurrent();
	double getUpTotal();
	double getUpCurrent();

	/**
	 * Aborts the transfer if it is still running. Note that this function will
	 * return immediately even though aborting may take some time. "wait" or
	 * "getPerforming" can be used to control this. The destructor will call
	 * this method automatically.
	 */
	void abort() const;

	/** Waits for the transfer to finish. */
	void wait() const;

	/** Returns whether the transfer is still in progress. */
	bool getPerforming() const;

	/** When finished it returns whether the transfer was successful. */
	bool getSuccessful() const;

	/** Returns the http response code if the transfer was successful. */
	long getHTTPCode() const;

private:

	/**
	 * Holds data that is shared between the thread and this transfer class.
	 */
	class SharedData {

	public:

		SharedData();

	private:

		struct Progress {
			Progress() : dltotal(0), dlnow(0), ultotal(0), ulnow(0) {}
			double dltotal;
			double dlnow;
			double ultotal;
			double ulnow;
		};


		/** Holds the current progress. */
		SharedObject<Progress> progress;

		/** Abortion flag set by the transfers destructor. */
		SharedObject<bool> aborting;

		/** Stores the http code when the transfer was successful. */
		SharedObject<long> httpCode;

		friend class HTTPTransfer;

	};


	/**
	 * Runner for the thread used for the transfer.
	 */
	class Runner : public Thread::Runner {

	public:

		Runner(const HTTPRequest& request,
			const Thread::SharedPtr<SharedData>& shared);

		void run();
		
	private:

		/** Holding the setup for the request. */
		HTTPRequest request;

		/** Shared pointer to shared data. */
		Thread::SharedPtr<SharedData> shared;

		friend class HTTPTransfer;

	};


	Thread::SharedPtr<SharedData> shared;
	Thread thread;


	/**
	 * Callback used by cURL for updating the current progress. "clientp"
	 * is treated as a pointer to the runner class above.
	 */
	static int callbackProgress(void *clientp,
		double dltotal, double dlnow, double ultotal, double ulnow);

	/**
	 * Callback used by cURL as read callback. "userdata" is used as a
	 * pointer to an object of "Source::Reader".
	 */
	static size_t callbackRead(
		void *ptr, size_t size, size_t nmemb, void *userdata);

	/**
	 * Callback used by cURL as write callback. "userdata" is used as a
	 * pointer to an object of "Destination::Writer".
	 */
	static size_t callbackWrite(
		void *ptr, size_t size, size_t nmemb, void *userdata);


	/** Copying and assigning prohibited. */
	HTTPTransfer(const HTTPTransfer& obj);
	HTTPTransfer& operator=(const HTTPTransfer& rhs);

};

#endif