#ifndef SHARED_OBJECT_H
#define SHARED_OBJECT_H

#include <windows.h>


/**
 * Template for shared objects. It provides a mechanism for locking an
 * object of the given template type for exclusive access by the current
 * thread.
 */
template <typename T>
class SharedObject {

public:

	/**
	 * A shared object can not be accessed directly through a member of
	 * "SharedObject". Instead you must create an object of class "Lock" and
	 * pass the shared object to its constructor. Then the shared object is
	 * locked as long as the lock object exists and can be accessed through it.
	 */
	class Lock {
		
	public:
		
		Lock(SharedObject<T>& sobj);
		~Lock();

		T& get();

	private:

		SharedObject<T>& sobj;

	};


	SharedObject(const T& object = T());
	SharedObject(const SharedObject<T>& obj);
	~SharedObject();

	SharedObject& operator=(SharedObject<T> rhs);
	void swap(SharedObject<T>& rhs);

private:

	T object;
	CRITICAL_SECTION cs;

};



template <typename T>
SharedObject<T>::SharedObject(const T& object) : object(object) {
	InitializeCriticalSectionAndSpinCount(&cs,0);
}

template <typename T>
SharedObject<T>::SharedObject(const SharedObject<T>& obj) :
	object(obj.object) {

	InitializeCriticalSectionAndSpinCount(&cs,0);
}

template <typename T>
SharedObject<T>::~SharedObject() {
	DeleteCriticalSection(&cs);
}

template <typename T>
SharedObject<T>& SharedObject<T>::operator=(SharedObject<T> rhs) {
	swap(rhs);
	return *this;
}

template <typename T>
void SharedObject<T>::swap(SharedObject<T>& rhs) {
	object.swap(rhs.object);
	std::swap(cs,rhs.cs);
}

template <typename T>
SharedObject<T>::Lock::Lock(SharedObject<T>& sobj) : sobj(sobj) {
	EnterCriticalSection(&sobj.cs);
}

template <typename T>
SharedObject<T>::Lock::~Lock() {
	LeaveCriticalSection(&sobj.cs);
}

template <typename T>
T& SharedObject<T>::Lock::get() {
	return sobj.object;
}

#endif