#ifndef SHARED_PTR_H
#define SHARED_PTR_H

#include "sharedObject.h"


/**
 * Reference counter for shared pointer. Must be 0 on construction.
 */
class SharedPtrCounter;


/**
 * Represents a reference-counting smart pointer.
 */
template <typename T>
class SharedPtr {

public:

	/**
	 * Uses "SharedPtrCounterStd". Therefore, the passed object must have been
	 * created with the "new" operator.
	 */
	SharedPtr(T* ref);

	/**
	 * Both "ref" and "refNumb" must have been created with the "new" operator
	 * and are destroyed with the "delete" operator if the reference counter
	 * reaches 0.
	 */
	SharedPtr(T* ref, SharedPtrCounter* refNumb);

	template <typename U>
	SharedPtr(const SharedPtr<U>& obj);
	SharedPtr(const SharedPtr& obj);

	virtual ~SharedPtr();

	SharedPtr<T>& operator=(SharedPtr rhs);
	void swap(SharedPtr& rhs);

	T& operator*();
	T* operator->() const;
	operator T*() const;

private:

	T* ref;
	SharedPtrCounter* refNumb;

};


class SharedPtrCounter {

public:

	virtual ~SharedPtrCounter();

	/**
	 * Must increase the reference counter by 1.
	 */
	virtual void operator++() = 0;

	/**
	 * Must decrease the reference counter by 1 and return whether it
	 * reached 0.
	 */
	virtual bool operator--() = 0;

};


/**
 * Basic implementation with no thread safety.
 */
class SharedPtrCounterStd : public SharedPtrCounter {

public:

	/**
	 * The passed object will be destroyed automatically with the "delete"
	 * operator and therefore must have been created with the "new"
	 * operator.
	 */
	SharedPtrCounterStd();

	void operator++();
	bool operator--();

private:

	size_t count;

};



template <typename T>
SharedPtr<T>::SharedPtr(T* ref) try :
	ref(ref), refNumb(new SharedPtrCounterStd) {
	
	++(*refNumb);
} catch (...) {
	delete ref;
}

template <typename T>
SharedPtr<T>::SharedPtr(T* ref, SharedPtrCounter* refNumb) :
	ref(ref), refNumb(refNumb) {

	++(*refNumb);
}

template <typename T>
template <typename U>
SharedPtr<T>::SharedPtr(const SharedPtr<U>& obj) :
	ref(obj.ref), refNumb(obj.refNumb) {

	++(*refNumb);
}

template <typename T>
SharedPtr<T>::SharedPtr(const SharedPtr& obj) :
	ref(obj.ref), refNumb(obj.refNumb) {

	++(*refNumb);
}

template <typename T>
SharedPtr<T>::~SharedPtr() {
	if (--(*refNumb)) {
		delete ref;
		delete refNumb;
	}
}

template <typename T>
SharedPtr<T>& SharedPtr<T>::operator=(SharedPtr rhs) {
	swap(rhs);
	return *this;
}

template <typename T>
void SharedPtr<T>::swap(SharedPtr& rhs) {
	std::swap<T*>(ref,rhs.ref);
	std::swap<SharedPtrCounter*>(refNumb,rhs.refNumb);
}

template <typename T>
T& SharedPtr<T>::operator*() {
	return *ref;
}

template <typename T>
T* SharedPtr<T>::operator->() const {
	return ref;
}

template <typename T>
SharedPtr<T>::operator T*() const {
	return ref;
}

#endif