#ifndef SOURCE_H
#define SOURCE_H

#include <string>
#include <vector>
#include "derived.h"
#include "sharedPtr.h"
#include "file.h"


/**
 * An object of a derived class represents a data source. "data" in this
 * context simply refers to a sequence of bytes. Actually such a source just
 * defines the location of the data. To read it, an object of the nested class
 * "Reader" is needed. It can be retrieved by calling "getReader". The returned
 * object then provides members to operate on the source, including reading
 * data, rewinding and determining its size. The reader must be independent
 * from the source it has been obtained from, i.e. the source is allowed to be
 * destroyed even if a reader for it still exists.
 */
class Source {

public:

	/**
	 * Default exception class.
	 */
	class Exception {};


	/**
	 * Provides access to a source. Its constructor and destructor are supposed
	 * to open/connect or to close/disconnect, respectively.
	 */
	class Reader {

	public:

		virtual ~Reader();

		/**
		 * Reads at most "numb" bytes from the source and writes them to
		 * "data", which therefore must be able to store at least "numb" bytes.
		 * Returns the actual number of bytes read from the source and stored
		 * in "data". A value of 0 indicates the end of the source. It is
		 * supposed that the number of bytes that "read" will retrieve before
		 * it reaches the end matches with the return value of "size".
		 */
		virtual size_t read(char* data, size_t numb) = 0;

		/**
		 * Rewinds the source so that "read" starts from the beginning.
		 */
		virtual void rewind() = 0;

		/**
		 * Returns the total number of bytes the source provides. The default
		 * implementation simply reads the complete source using "read" and
		 * rewinds afterwards. This will always work but may be slow especially
		 * if the source provides a lot of data. Therefore, whenever it is
		 * possible to determine the size without a complete read through, this
		 * member should be overloaded.
		 */
		virtual size_t size();

		/**
		 * Reads the complete source and puts the data together in a string
		 * object, which is returned. Afterwards, the source is rewinded.
		 */
		std::string toString();

		/**
		 * Defines a size for buffers. It might be used whenever a buffer for
		 * the "read" member is needed.
		 */
		static const size_t BUFFER_SIZE = 1 << 16; // 64kB

	};


	virtual ~Source();
	
	/**
	 * Returns a reader object for this source.
	 */
	virtual Derived<Reader> getReader() = 0;

};


/**
 * Uses a buffer as source.
 */
class SourceBuffer : public Source {

public:

	/**
	 * Corresponding reader.
	 */
	class Reader : public Source::Reader {
	
	public:

		Reader(const std::string& buffer);

		size_t read(char* data, size_t numb);
		void rewind();
		size_t size();

	private:

		std::string buffer;
		unsigned int pos;

	};


	SourceBuffer(const std::string& buffer);

	Derived<Source::Reader> getReader();

private:

	std::string buffer;

};


/**
 * Uses a file as source.
 */
class SourceFile : public Source {
	
public:

	/**
	 * Corresponding reader. Note that copied readers refer to the same file
	 * handle and therefore will share the same position indicator.
	 */
	class Reader : public Source::Reader {
		
	public:

		Reader(const std::string& file);

		size_t read(char* data, size_t numb);
		void rewind();
		size_t size();

	private:

		SharedPtr<File> handle;

	};


	SourceFile(const std::string& file);

	Derived<Source::Reader> getReader();

private:

	std::string file;

};


/**
 * Concatenates other sources.
 */
class SourceConcat : public Source {

public:

	/**
	 * Corresponding reader.
	 */
	class Reader : public Source::Reader {
		
	public:

		Reader(const std::vector<Derived<Source> >& src);

		size_t read(char* data, size_t numb);
		void rewind();
		size_t size();

	private:

		std::vector<Derived<Source::Reader> > reader;
		unsigned int current;

	};


	SourceConcat(const std::vector<Derived<Source> >& src);

	Derived<Source::Reader> getReader();

private:

	std::vector<Derived<Source> > src;

};

#endif