#ifndef THREAD_H
#define THREAD_H

#include <windows.h>
#include "derived.h"
#include "sharedPtr.h"
#include "sharedObject.h"


/**
 * An object of this class represents a single thread of the OS. It is started
 * with its construction but its destruction does not automatically terminate
 * the thread and is not even affected by it. The class uses the windows api
 * for implementation so it will only compile on windows.
 */
class Thread {

public:

	/** Exception type for this class. */
	class Exception {};


	/**
	 * Thread safe shared pointer.
	 */
	template <typename T>
	class SharedPtr : public ::SharedPtr<T> {

	public:

		/**
		 * The passed object is automatically destroyed with the "delete"
		 * operator when all shared pointers to it are gone. Therefore, the
		 * object must have been created with the "new" operator.
		 */
		SharedPtr(T* ref);
		SharedPtr(const SharedPtr& obj);

		SharedPtr<T>& operator=(const SharedPtr& rhs);

	};


	/**
	 * Corresponding counter.
	 */
	class SharedPtrCounter : public ::SharedPtrCounter {

	public:

		SharedPtrCounter();

		void operator++();
		bool operator--();

	private:

		SharedObject<size_t> count;

	};


	/**
	 * A new thread uses "run" as entry point. It may throw any exception to
	 * indicate an error.
	 */
	class Runner {

	public:

		virtual void run() = 0;

	};


	/**
	 * Starts a new thread using the given runner, i.e. calls its ()-operator.
	 * If it throws an exception the thread is terminated and the "successful"
	 * member will return false. The given shared pointer to the runner is
	 * copied in the constructor and destroyed when the thread ends, i.e. after
	 * the ()-operator returned or threw.
	 */
	Thread(const Derived<Runner>& runner);
	~Thread();

	/**
	 * Waits until the thread terminated, but at most "timeout" milliseconds.
	 * You may pass 0 to just check whether the thread already terminated
	 * without waiting or omit the parameter for no timeout. It returns "true"
	 * if the timeout expired and "false" if the thread terminated before.
	 */
	bool wait(unsigned long timeout = INFINITE) const;

	/**
	 * Returns whether the thread terminated successfully, i.e. whether the
	 * procedure passed to the constructor returned normally without throwing
	 * an exception. It returns false if the thread is still running.
	 */
	bool successful() const;

private:

	HANDLE handle;


	/** Copying and assigning prohibited. */
	Thread(const Thread& obj);
	Thread& operator=(const Thread& obj);


	/**
	 * Used as entry point wrapper. Is called by a new thread and calls the
	 * procedure passed to the constructor.
	 */
	static DWORD WINAPI procWrapper(LPVOID param);

};



template <typename T>
Thread::SharedPtr<T>::SharedPtr(T* ref) try :
	::SharedPtr<T>(ref,new SharedPtrCounter) {}
	catch (...) {
		delete ref;
	}

template <typename T>
Thread::SharedPtr<T>::SharedPtr(const SharedPtr& obj) :
	::SharedPtr<T>(obj) {}

template <typename T>
Thread::SharedPtr<T>& Thread::SharedPtr<T>::operator=(const SharedPtr& rhs) {
	::SharedPtr<T>::operator=(rhs);
	return *this;
}

#endif